package algae.simulation;

import java.util.*;

import algae.simulation.Settings;
import algae.simulation.worlds.*;
import algae.simulation.worlds.PheromoneWorld;
import algae.simulation.worlds.SubstrateWorld; 
import algae.simulation.agents.*;
import algae.simulation.SimulationStats;

// worldcontroller for batch mode
public class BatchWC extends WC {

    boolean initialized=false;

    public String fileName;

    public BatchWC(Settings set, String fn) {
	super(set);
	fileName=fn;
	initialized=true;
    }
    
    public void run() {
    	WorldState ws;
    	
    	if (!initialized) 
    	{
    	    System.out.println("Initialization of BatchController failed...");
    	    return;
    	}
    	
    	SimulationStats stats=new SimulationStats(fileName);
	if (!stats.openStream(cfg,agents))
	System.out.println("IOERROR: cannot create stream");
	
    	while (stepsRunning<stepsToRun)
    	{
	    step();
	
	    if (!stats.writeStats(stepsRunning))
	       System.out.println("IOERROR: cannot write to stream");
		    
	    ws=new WorldState(this); // save it for statistics
	    System.out.println(stepsRunning+" stepdone (of "+stepsToRun+" steps)");
	}
	System.out.println("\nSimulation finished after "+stepsRunning+" step(s)\n");
	System.out.println("Writing simulation-summary into statistics file now.");
	System.out.println("Please be patient - this may take some time...\n");
	
	if (!stats.writeSummary())
	   System.out.println("IOERROR: cannot write summary");
	else
	   System.out.println("Summary saved successfully...");
	 
	
	if (!stats.closeStream())
	   System.out.println("IOERROR: cannot close stream");
	else
	   System.out.println("Statistics saved successfully...");
    }
    
}
