package algae.simulation.agents;

import algae.simulation.worlds.AgentWorld;
import algae.simulation.worlds.LightWorld;
import algae.simulation.worlds.SubstrateWorld;
import algae.simulation.worlds.PheromoneWorld;
import algae.simulation.worlds.World;
import algae.simulation.Settings;

import java.io.IOException;

import java.lang.Math;
import java.util.*;

public class AgentSensor {
	
	public int range=0;
	public int fov=0;
	Vector visited=new Vector(5,1); 
	 
	int visitCount;
	 
	Settings cfg;
	
	public AgentSensor(Settings c, int iRange, int iFov) 
	{
	     cfg=c;
	     fov=iFov;
	     range=iRange;
	}	
	
	public boolean inBounds(int x, int y)
	{
	     if ((x>=0)&&(x<cfg.width)&&(y>=0)&&(y<cfg.height))
	     	return true;
	     else
	     	return false;
	}
	
	// was position x/y already tested (searched) by a specific
	// sensor
	public boolean alreadyVisited(int x, int y)
	{
	     int testX=0;
	     int testY=0;
	     
	     for (int z=0; z<visitCount; z++)
	     {
	     	Integer tmp=((Integer)visited.elementAt(z*2)); 	  
	     	testX=tmp.intValue();
	     	
	     	tmp=((Integer)visited.elementAt(z*2+1)); 	  
	     	testY=tmp.intValue();
	     	
	     	if ((testY==y)&&(testX==x))
	     	   return true;
	     }
	     
	     return false;
	}
	
	// getValues() for BumperSensor
	public float[] getValuesA(int d, int x, int y, AgentWorld w)
	{
	   float retVal[]=new float[4];
		
	   for (int dir=0; dir<4; dir++)
	   {
	      visited.clear();	
	      visitCount=0;
	   	
              int dd=dir+d;
              if (dd>=4)
                 dd=dd-4;
	      retVal[dir]=0;
	      for (int r=0; r<range; r++)
	      {
	         for (int angle=-fov/2; angle<fov/2; angle++)
	     	 {
	     	     double sa=angle+dd*90;
	     	     if (sa<0.0)
	     	     	sa=sa+360.0;
	     	     if (sa>360.0)
	     	     	sa=sa-360.0;
	     	     	
	     	     int sx=x+(int)((double)(r+1)*Math.cos(sa/180.0*Math.PI));
	     	     int sy=y+(int)((double)(r+1)*Math.sin(sa/180.0*Math.PI));
	     	     
	     	     if (alreadyVisited(sx,sy))
	     	        continue;
	     	        
	     	     visited.add(new Integer(sx));
	     	     visited.add(new Integer(sy));
	     	     visitCount++;
	     	     
	     	     if (inBounds(sx,sy))
	     	     {
	     	     	if (w.world[sy][sx].amount!=0.0)
	     	     	   retVal[dir]=retVal[dir]+0.1f;
	     	     }
	     	     else
	     	     {
	     	     	retVal[dir]=retVal[dir]+10.0f;
           	     }
	     	 }
	      }     
	   }
	   return retVal;
	}
	
	// getValues() for PheromoneSensor
	public float[] getValuesP(int d, int x, int y, PheromoneWorld w)
	{
	   float retVal[]=new float[4];
	   	   
	   for (int dir=0; dir<4; dir++)
	   {
	      visited.clear();
	      visitCount=0;
	      
              int dd=dir+d;
              if (dd>=4)
                 dd=dd-4;
	      retVal[dir]=0;
	      for (int r=0; r<range; r++)
	      {
	         for (int angle=-fov/2; angle<fov/2; angle++)
	     	 {
	     	     double sa=angle+dd*90;
	     	     if (sa<0.0)
	     	     	sa=sa+360.0;
	     	     if (sa>360.0)
	     	     	sa=sa-360.0;
	     	     	
	     	     int sx=x+(int)((double)(r+1)*Math.cos(sa/180.0*Math.PI));
	     	     int sy=y+(int)((double)(r+1)*Math.sin(sa/180.0*Math.PI));
	     	     
	     	     if (alreadyVisited(sx,sy))
	     	        continue;
	     	        
	     	     visited.add(new Integer(sx));
	     	     visited.add(new Integer(sy));
	     	     visitCount++;
	     	     
	     	     if (inBounds(sx,sy))
	     	     {
	     	     	retVal[dir]=retVal[dir]+(float)w.world[sy][sx].amount;
	     	     }	             
	     	 }
	      }     
	   }
	   
	   return retVal;
	}

	// getValues() for SubstrateSensor
	public float[] getValuesS(int d, int x, int y, SubstrateWorld w)
	{
	   float retVal[]=new float[4];
	  	   	
	   for (int dir=0; dir<4; dir++)
	   {
	      visited.clear();
	      visitCount=0; 

	      int dd=dir+d;
              if (dd>=4)
                 dd=dd-4; 
	      retVal[dir]=0;
	      for (int r=0; r<range; r++)
	      {
	         for (int angle=-fov/2; angle<fov/2; angle++)
	     	 {
	     	     
	     	     double sa=angle+dd*90;
	     	     if (sa<0.0)
	     	     	sa=sa+360.0;
	     	     if (sa>360.0)
	     	     	sa=sa-360.0;
	     	     
	     	     int sx=x+(int)((double)(r+1)*Math.cos(sa/180.0*Math.PI));
	     	     int sy=y+(int)((double)(r+1)*Math.sin(sa/180.0*Math.PI));
	     
	     	     if (alreadyVisited(sx,sy))
	     	        continue;
	     	        
	     	     visited.add(new Integer(sx));
	     	     visited.add(new Integer(sy));
	     	     visitCount++;
	     	     	     	     
	     	     if (inBounds(sx,sy))
	     	     {
	     	     	retVal[dir]=retVal[dir]+(float)w.world[sy][sx].amount;
	     	     }	     	     
	     	 }
	      }    
	   }
	   return retVal;
	}
	
	// getValues() for LightSensor
	public float[] getValuesL(int d, int x, int y, LightWorld w)
	{
	   float retVal[]=new float[4];
	  	   	
	   for (int dir=0; dir<4; dir++)
	   {
	      visited.clear();
	      visitCount=0; 
	   	
	      int dd=dir+d;
	      if (dd>=4)
                 dd=dd-4;
	      retVal[dir]=0;
	      for (int r=0; r<range; r++)
	      {
	         for (int angle=-fov/2; angle<fov/2; angle++)
	     	 {
	     	     double sa=angle+dd*90;
	     	     if (sa<0.0)
	     	     	sa=sa+360.0;
	     	     if (sa>360.0)
	     	     	sa=sa-360.0;
	     	     
	     	     int sx=x+(int)((double)(r+1)*Math.cos(sa/180.0*Math.PI));
	     	     int sy=y+(int)((double)(r+1)*Math.sin(sa/180.0*Math.PI));
	     	     
	     	     if (alreadyVisited(sx,sy))
	     	        continue;
	     	        
	     	     visited.add(new Integer(sx));
	     	     visited.add(new Integer(sy));
	     	     visitCount++;
	     	     	     	     
	     	     if (inBounds(sx,sy))
	     	     {
	     	     	retVal[dir]=retVal[dir]+(float)w.world[sy][sx].amount;
	     	     }	     	     
	     	 }
	      }    
	   }
	   return retVal;
	}

}
