package algae.simulation.agents;

import java.util.*;
import java.lang.Object;
import java.lang.*;
import java.util.BitSet;
import java.util.Random;
import java.lang.Math;

public class GenoType {

    // the DNA string
    BitSet nm_bit = new BitSet(2500);  // 50x50 "booleans" 
    BitSet wm_bit = new BitSet(80000); // 50x50 32-bit IEEE floats
    BitSet sm_bit = new BitSet(640);   // 10x2 32-bit IEEE floats
           
    Random rand;
    
    int crippled = 0;
    int randSeed=0;
    
    // constructor for randomly initialized agent
    public GenoType(int rs)
    {
    	randSeed=rs;
    	rand=new Random(randSeed);
    	initRandomNm();
    	initRandomWm();
    	initRandomSm();
    }
    
    // constructor for cross over agent (meiosis)
    public GenoType(int rs, GenoType a, GenoType b, float crossOverRate, float mutationRate)
    {
    	randSeed=rs;
    	rand=new Random(randSeed);
    	this.crossOver(a,b,crossOverRate,mutationRate);
    	this.mutate(mutationRate);
    }
    
    // constructor for mutation (mitosis)
    public GenoType(int rs, GenoType a, float mutationRate)
    {
    	rand=new Random(randSeed);
    	randSeed=rs;
    	this.sm_bit=a.sm_bit;
    	this.nm_bit=a.nm_bit;
    	this.wm_bit=a.wm_bit;
    	this.mutate(mutationRate);
    }
    
    // return the crippleness of an agent (after mutation and/or crossover)
    public int crippleness()
    {
	return crippled;
    }

    // return if agent is crippled
    public boolean isCrippled()
    {
    	return (crippled>0);
    }
    
    // make a new genotype out of two others (meiosis)
    public void crossOver(GenoType a, GenoType b, 
    		float crossOverRate, float mutationRate)
    {
    	int crossingPos = rand.nextInt() % nm_bit.size();
    	
    	int cs=0;
    	
    	// crossing over nm_bit
    	for (int z=0; z<a.nm_bit.size(); z++)
    	{
    	    if (rand.nextFloat()<crossOverRate)
    	    	cs=1-cs;
    		
    	    if (cs==0)
    	    {
    	    	if (b.nm_bit.get(z)==true)
    	    	    nm_bit.set(z);
    	    	else
    	    	    nm_bit.clear(z);
    	    }
    	    else
    	    {
    	    	if (a.nm_bit.get(z)==true)
    	    	    nm_bit.set(z);
    	    	else
    	    	    nm_bit.clear(z);
    	    }
    	}
    	
    	// validate the nm_bit matrix
	for (int i=0; i<50; i++)
	    for (int j=0; j<50; j++)
	    	if ((i>j)&&(nm_bit.get(j*50+i)==true))
	    	{
	    	   nm_bit.clear(j*50+i);
	    	}
    
    	// crossing over wm_bit
    	for (int z=0; z<a.wm_bit.size(); z++)
    	{
    	    if (rand.nextFloat()<crossOverRate)
    	    	cs=1-cs;
    		
    	    if (cs==0)
    	    {
    	    	if (b.wm_bit.get(z)==true)
    	    	    wm_bit.set(z);
    	    	else
    	    	    wm_bit.clear(z);
    	    }
    	    else
    	    {
    	    	if (a.wm_bit.get(z)==true)
    	    	    wm_bit.set(z);
    	    	else
    	    	    wm_bit.clear(z);
    	    }
    	}
    	
    	// don't cross over spline_matrix -> next version
    	for (int z=0; z<a.sm_bit.size(); z++)
    	{
    	     if (a.sm_bit.get(z)==true)
    	         sm_bit.set(z);
    	     else
    	         sm_bit.clear(z);	  
    	}
    	
    	/*
    	// crossing over sm_bit
    	for (int z=0; z<a.sm_bit.size(); z++)
    	{
    	    if (rand.nextFloat()<crossOverRate)
    	    	cs=1-cs;
    		
    	    if (cs==0)
    	    {
    	    	if (b.sm_bit.get(z)==true)
    	    	    sm_bit.set(z);
    	    	else
    	    	    sm_bit.clear(z);
    	    }
    	    else
    	    {
    	    	if (a.sm_bit.get(z)==true)
    	    	    sm_bit.set(z);
    	    	else
    	    	    sm_bit.clear(z);
    	    }
    	}
    	
    	for (int z=0; z<10-1; z++)
    	{
    	     // TODO: splines' x-values are mixed up
    	     if (getFloatOuttaSm(z*2)>getFloatOuttaSm((z+1)*2))
    	     	crippled++;
    	}
    	*/
    }

    // mutation of the genotype (mitosis)
    public void mutate(float mutationRate)
    {
    	// let the mutation begin :)
    	// first we mutate the wm_bit matrix    	
    	for (int z=0; z<nm_bit.size(); z++)
    	{
    	    if (rand.nextFloat()<mutationRate)
    	    {
    	    	if (nm_bit.get(z)==true)
    	    	   nm_bit.clear(z);
    	    	else
    	    	   nm_bit.set(z);
    	    }
    	}
    	
    	// validate the nm_bit matrix
	for (int i=0; i<50; i++)
	    for (int j=0; j<50; j++)
	    	if ((i>j)&&(nm_bit.get(j*50+i)==true))
	    	{
	    	   nm_bit.clear(j*50+i);
	    	}
	
	// now let's go to weight_matrix mutation    	
	for (int z=0; z<wm_bit.size(); z++)
    	{
    	    if (rand.nextFloat()<mutationRate)
    	    {
    	    	if (wm_bit.get(z)==true)
    	    	   wm_bit.clear(z);
    	    	else
    	    	   wm_bit.set(z);
    	    }
    	}
    	
    	// check for validity of IEEE numbers
    	for (int z=0; z<50*50; z++)
    	{
    	    // first make all signed values to unsigned ones
    	    if (wm_bit.get(z*32)==true)
    	    	wm_bit.clear(z*32);
    	    
    	    // check if floats are valid	
    	    if (getFloatOuttaWm(z)==Float.NaN)
    	    {
    	    	crippled++;
    	    }
    	}
    	    	    	
    	/*
    	// and last, mutate the spline matrix
    	for (int z=0; z<sm_bit.size(); z++)
    	{
    	    if (rand.nextFloat()<mutationRate)
    	    {
    	    	if (sm_bit.get(z)==true)
    	    	   sm_bit.clear(z);
    	    	else
    	    	   sm_bit.set(z);
    	    }
    	}
    	
    	// and check out if the splines are "good"
    	for (int z=0; z<10; z++)
    	{	
    	    // check if floats are valid 	
    	    if (getFloatOuttaSm(z*2)==Float.NaN)
    	    {
    	    	// TODO: decide how we should handle this
    	    	crippled++;
    	    }
    	}
    	
    	// check the splines for correct x-values (x1<x2<...<xn)
    	for (int z=0; z<10-1; z++)
    	{
    	    if (getFloatOuttaSm(z*2)>getFloatOuttaSm((z+1)*2))
    	    {
    	    	// got a bad one
    	    	// TODO: decide how we should handle this
    	    	crippled++;
    	    }
    	}
    	*/
    }

    // convert a bitset (32 bit length) to an IEEE float
    public float IEEE2float(BitSet f)
    {
    	int z=0;
    	float sign=0.0f;    	
    	float exponent=0.0f;
    	float mantissa=0.0f;
    	
    	// check for signed float
    	if (f.get(31)==true)
    	   sign=-1.0f;
    	else
           sign=1.0f; 		
           
        // extract exponent out of bitstring
        for (z=1; z<9; z++)
        {
           float digit=0.0f;
           if (f.get(31-z)==true)
               digit=1.0f; 	
           exponent=exponent+digit*(float)Math.pow((double)2,(double)(8-z));
        }
        
        if (exponent==255.0f)
            return Float.NaN;
        
        exponent=exponent-127.0f;
        
        // extract mantissa out of bitstring
        for (z=9; z<32; z++)
        {
           float digit=0.0f;
           if (f.get(31-z)==true)
               digit=1.0f;
           mantissa=mantissa+digit*1.0f/((float)Math.pow((double)2,(double)(z-8)));    
        }
        mantissa=mantissa+1;
        
        // put all the stuff together
        return sign*mantissa*(float)Math.pow((double)2,(double)exponent);
    }

    // get a bitset out of a 32 bit IEEE float
    public BitSet float2IEEE(float f)
    {
    	BitSet b=new BitSet(32);
    	    	    	
    	// get integer corresponding to our 32 bit IEEE bitstring 
    	int ln=Float.floatToIntBits(f);
    	
    	int z=0;
    	// convert to bitstring
    	while (ln>0)
    	{
    	   if (ln%2==1)
    	       b.set(z);
	   ln=ln/2;   
	   z++; 	    	    
    	}
    	 
    	return b;   	
    }
    
    // get a single weight value out of the weight_matrix bitstring
    public float getFloatOuttaWm(int idx)
    {
    	BitSet tmp=new BitSet(32);
    	for (int z=0; z<32; z++)
    	     if (wm_bit.get(idx*32+z)==true)
    	     	tmp.set(z);
    	     else
    	     	tmp.clear(z);
    	     	
    	return IEEE2float(tmp);
    }

    // get a single spline (x/y) value out of the spline_matrix bitstring 
    public float getFloatOuttaSm(int idx)
    {
    	BitSet tmp=new BitSet(32);
    	for (int z=0; z<32; z++)
    	     if (sm_bit.get(idx*32+z)==true)
    	     	tmp.set(z);
    	     else
    	     	tmp.clear(z);
    	     	
    	return IEEE2float(tmp);
    }
    
    // store a value into weight_matrix at position idx
    public void putFloatIntoWm(int idx, float f)
    {
    	BitSet tmp=float2IEEE(f);
    	for (int z=0; z<32; z++)
    	{
    	     if (tmp.get(z)==true)
    	     	wm_bit.set(32*idx+z);
    	     else
    	     	wm_bit.clear(32*idx+z);
    	}
    }
    
    // store a value into spline_matrix at position idx
    public void putFloatIntoSm(int idx, float f)
    {
    	BitSet tmp=float2IEEE(f);
    	for (int z=0; z<32; z++)
    	{
    	     if (tmp.get(z)==true)
    	     	sm_bit.set(32*idx+z);
    	     else
    	     	sm_bit.clear(32*idx+z);
    	}
    }
        
    // initalize the weight_matrix with random values
    public void initRandomWm()
    {
	for (int z=0; z<50*50; z++)
	    putFloatIntoWm(z,rand.nextFloat());
    }
    
    // initialize the neuron_matrix (MMM) with random value and check them for correctness
    public void initRandomNm()
    {
	for (int z=0; z<50*50; z++)
	    if (rand.nextFloat()>0.5f)
		nm_bit.set(z);
		
	// validate the matrix
	for (int i=0; i<50; i++)
	    for (int j=0; j<50; j++)
	    	if ((i>j)&&(nm_bit.get(j*50+i)==true))
	    	   nm_bit.clear(j*50+i);
    }
    
    // initialize the spline_matrix with random values
    public void initRandomSm()
    {
	for (int z=0; z<10; z++)
	{
	    putFloatIntoSm(z*2,(float)(z*20));
	    putFloatIntoSm(z*2+1,rand.nextFloat());
	}		
    }
    
    // show the bitstring for a IEEE float
    public void showIEEE32(BitSet b)
    {
    	for (int z=0; z<32; z++)
    	    if (b.get(31-z)==true)
    	    	System.out.print("1");
    	    else
    	    	System.out.print("0");
    	System.out.println("");    	     	 
    }
    
    // pretty much the same as showIEEE32
    public void showIEEE(BitSet f)
    {
    	System.out.println(f.toString());
    }
}
